import asyncio
import logging
from datetime import datetime, timedelta, timezone
from typing import Awaitable, Callable, List, Tuple

from golem.managers.base import DemandManager
from golem.managers.mixins import BackgroundLoopMixin
from golem.node import GolemNode
from golem.payload import Payload
from golem.payload import defaults as payload_defaults
from golem.resources import Allocation, Demand, Proposal
from golem.resources.demand.demand_builder import DemandBuilder
from golem.utils.asyncio import ErrorReportingQueue, create_task_with_logging, ensure_cancelled_many
from golem.utils.logging import get_trace_id_name, trace_span

logger = logging.getLogger(__name__)


class RefreshingDemandManager(BackgroundLoopMixin, DemandManager):
    def __init__(
        self,
        golem: GolemNode,
        get_allocation: Callable[[], Awaitable[Allocation]],
        payloads: List[Payload],
        demand_lifetime: timedelta = payload_defaults.DEFAULT_LIFETIME,
        subnet_tag: str = payload_defaults.DEFAULT_SUBNET,
    ) -> None:
        self._golem = golem
        self._get_allocation = get_allocation
        self._payloads = payloads
        self._demand_lifetime = demand_lifetime
        self._subnet_tag = subnet_tag

        self._initial_proposals: ErrorReportingQueue[Proposal] = ErrorReportingQueue()

        self._demands: List[Tuple[Demand, asyncio.Task]] = []
        super().__init__()

    @trace_span("Starting RefreshingDemandManager", log_level=logging.INFO)
    async def start(self) -> None:
        return await super().start()

    @trace_span("Stopping RefreshingDemandManager", log_level=logging.INFO)
    async def stop(self) -> None:
        return await super().stop()

    @trace_span("Getting initial proposal", show_results=True)
    async def get_initial_proposal(self) -> Proposal:
        proposal = await self._initial_proposals.get()
        self._initial_proposals.task_done()
        return proposal

    @trace_span()
    async def _background_loop(self) -> None:
        try:
            while True:
                if self._demands:
                    await self._wait_for_demand_to_expire()
                    await self._stop_consuming_initial_proposals()

                await self._create_and_subscribe_demand()
        except Exception as e:
            self._initial_proposals.set_exception(e)
            logger.debug(
                "Encountered unexpected exception while handling demands,"
                " exception is set and background loop will be stopped!"
            )
        finally:
            await self._stop_consuming_initial_proposals()
            await self._unsubscribe_demands()

    async def _wait_for_demand_to_expire(self):
        await self._demands[-1][0].get_data()
        expiration_date = await self._demands[-1][0].get_expiration_date()
        remaining = expiration_date - datetime.now(timezone.utc)
        await asyncio.sleep(remaining.total_seconds())

    @trace_span()
    async def _create_and_subscribe_demand(self):
        allocation = await self._get_allocation()
        demand_builder = await self._prepare_demand_builder(allocation)
        logger.debug(f"Creating demand: {demand_builder=}")
        demand = await demand_builder.create_demand(self._golem)
        demand.start_collecting_events()
        await demand.get_data()
        self._demands.append(
            (
                demand,
                create_task_with_logging(
                    self._consume_initial_proposals(demand),
                    trace_id=get_trace_id_name(self, f"demand-{demand.id}-proposal-consumer-loop"),
                ),
            )
        )

    @trace_span()
    async def _consume_initial_proposals(self, demand: Demand):
        try:
            initial_proposals_gen = demand.initial_proposals()
            first_initial_proposal = await initial_proposals_gen.__anext__()
            logger.info("Received first initial proposal")

            logger.debug(f"New initial proposal {first_initial_proposal}")
            self._initial_proposals.put_nowait(first_initial_proposal)

            async for initial in initial_proposals_gen:
                logger.debug(f"New initial proposal {initial}")
                self._initial_proposals.put_nowait(initial)
        except asyncio.CancelledError:
            pass

    @trace_span()
    async def _stop_consuming_initial_proposals(self) -> None:
        await ensure_cancelled_many([d[1] for d in self._demands])

    @trace_span()
    async def _prepare_demand_builder(self, allocation: Allocation) -> DemandBuilder:
        # FIXME: Code looks duplicated as GolemNode.create_demand does the same
        demand_builder = DemandBuilder()

        for demand_spec in [
            payload_defaults.ActivityInfo(lifetime=self._demand_lifetime, multi_activity=True),
            payload_defaults.NodeInfo(subnet_tag=self._subnet_tag),
            payload_defaults.PaymentInfo(),
            await allocation.get_demand_spec(),
        ] + self._payloads:
            await demand_builder.add(demand_spec)

        return demand_builder

    @trace_span()
    async def _unsubscribe_demands(self):
        results = await asyncio.gather(
            *[demand.unsubscribe() for demand, _ in self._demands], return_exceptions=True
        )
        for result in results:
            if isinstance(result, Exception):
                logger.warning(f"Unable to unsubscribe demand due to {type(result)}:\n{result}")



from golem.exceptions import GolemException


class InputStreamExhausted(GolemException):
    """Excepion used internally by pipeline-level components.

    This is something like StopAsyncIteration, but in context when StopAsyncIteration can't be used.
    All mid-level components understand & correctly process this exception, the only way to
    encounter it in non-chain code is to create a Chain that returns awaitables that can fail
    because of the input stream that did not return enough values.

    E.g. this will raise InputStreamExhausted with 0.64 chance::

        async def source():
            yield 1
            yield 2

        async def flaky_processing_function(val):
            from random import random
            if random() > 0.2:
                raise Exception("Ooops")

            return val

        async for awaitable_result in Chain(
            source(),
            Map(flaky_processing_function),
        ):
            print(await awaitable_result)
            break

    But if we add a Buffer in the end of the chain::

        async for result in Chain(
            source(),
            Map(flaky_processing_function),
            Buffer(),
        ):
            print(result)
            break

    then this exception will be handled by the Buffer (and we have 0.64 chance anything will be
    printed).

    """



from typing import TYPE_CHECKING

from golem.exceptions import GolemException

if TYPE_CHECKING:
    from golem.resources.base import Resource


class ResourceException(GolemException):
    pass


class ResourceNotFound(ResourceException):
    """Raised on an attempt to interact with a resource that doesn't exist.

    Example::

        async with GolemNode() as golem:
            agreement_id = "".join(["a" for x in range(64)])
            agreement = golem.agreement(agreement_id)
            try:
                await agreement.get_data()
            except ResourceNotFound as e:
                print(f"Agreement with id {e.resource.id} doesn't exist")


    """

    def __init__(self, resource: "Resource"):
        self._resource = resource

        msg = f"{resource} doesn't exist"
        super().__init__(msg)

    @property
    def resource(self) -> "Resource":
        """Resource that caused the exception."""
        return self._resource


class PaymentValidationException(Exception):
    ...



from pathlib import Path
from typing import Optional

from textx import TextXSyntaxError, metamodel_from_file

from golem.payload.constraints import Constraint, ConstraintGroup, Constraints


class SyntaxException(Exception):
    pass


class PayloadSyntaxParser:
    __instance: Optional["PayloadSyntaxParser"] = None

    @classmethod
    def get_instance(cls):
        if not cls.__instance:
            cls.__instance = cls()
        return cls.__instance

    def __init__(self):
        self._metamodel = metamodel_from_file(str(Path(__file__).with_name("parser.tx")))
        self._metamodel.register_obj_processors(
            {
                "ConstraintGroup": lambda e: ConstraintGroup(e.items, e.operator),
                "Constraint": lambda e: Constraint(e.property_path, e.operator, e.value),
                "PropertyValueList": lambda e: e.items,
            }
        )

    def parse_constraints(self, syntax: str) -> Constraints:
        try:
            model = self._metamodel.model_from_str(syntax)
        except TextXSyntaxError as e:
            raise SyntaxException(f"Syntax `{syntax}` parsed with following error: {e}")

        return model.constraints



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceClosed, ResourceDataChanged

if TYPE_CHECKING:
    from golem.resources.allocation.allocation import Allocation  # noqa


class NewAllocation(NewResource["Allocation"]):
    pass


class AllocationDataChanged(ResourceDataChanged["Allocation"]):
    pass


class AllocationClosed(ResourceClosed["Allocation"]):
    pass



from golem.resources.debit_note.debit_note import DebitNote
from golem.resources.debit_note.event_collectors import DebitNoteEventCollector
from golem.resources.debit_note.events import DebitNoteClosed, DebitNoteDataChanged, NewDebitNote

__all__ = (
    "DebitNote",
    "DebitNoteEventCollector",
    "NewDebitNote",
    "DebitNoteDataChanged",
    "DebitNoteClosed",
)



from golem.resources.exceptions import ResourceException


class AllocationException(ResourceException):
    pass


class NoMatchingAccount(AllocationException):
    """Raised when a new :any:`Allocation` is created for a (network, driver) pair without \
    matching `yagna` account."""

    def __init__(self, network: str, driver: str):
        self._network = network
        self._driver = driver

        #   NOTE: we don't really care about this sort of compatibility, but this is
        #         a message developers are used to so maybe it's worth reusing
        msg = f"No payment account available for driver `{driver}` and network `{network}`"
        super().__init__(msg)

    @property
    def network(self) -> str:
        return self._network

    @property
    def driver(self) -> str:
        return self._driver



from typing import TYPE_CHECKING

from golem.resources.events import NewResource, ResourceEvent

if TYPE_CHECKING:
    from golem.resources.pooling_batch.pooling_batch import PoolingBatch  # noqa


class NewPoolingBatch(NewResource["PoolingBatch"]):
    pass


class BatchFinished(ResourceEvent["PoolingBatch"]):
    """Emitted when the execution of a :any:`PoolingBatch` finishes.

    The same event is emitted for successful and failed batches.
    """



import logging
from typing import Set

from golem.managers import ProposalManagerPlugin
from golem.resources import Proposal
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class WhitelistProviderIdPlugin(ProposalManagerPlugin):
    def __init__(self, whitelist: Set[str]) -> None:
        self._whitelist = whitelist

    @trace_span(show_results=True)
    async def get_proposal(self) -> Proposal:
        while True:
            proposal: Proposal = await self._get_proposal()
            proposal_data = await proposal.get_data()
            provider_id = proposal_data.issuer_id

            if provider_id in self._whitelist:
                return proposal

            if not proposal.initial:
                await proposal.reject("provider_id is not on whitelist")

            logger.debug(
                f"Provider id `{provider_id}` from proposal `{proposal}` is not on whitelist,"
                f" picking different proposal..."
            )



from golem.resources.activity import Activity
from golem.resources.agreement.agreement import Agreement


async def default_create_activity(agreement: Agreement) -> Activity:
    """Create a new :any:`Activity` for a given :any:`Agreement`."""
    return await agreement.create_activity()



import asyncio
from typing import Optional

import click

from golem.cli.utils import (
    CliPayload,
    async_golem_wrapper,
    format_allocations,
    format_demands,
    format_proposals,
    parse_timedelta_str,
)
from golem.node import GolemNode
from golem.payload.defaults import DEFAULT_PAYMENT_DRIVER, DEFAULT_PAYMENT_NETWORK, DEFAULT_SUBNET


@click.group()
def cli() -> None:
    pass


@cli.group()
def allocation() -> None:
    pass


@allocation.command("list")
@async_golem_wrapper
async def allocation_list(golem: GolemNode) -> None:
    allocations = await golem.allocations()
    click.echo(format_allocations(allocations))


@allocation.command("new")
@click.argument("amount", type=float)
@click.option("--network", type=str, default=DEFAULT_PAYMENT_NETWORK)
@click.option("--driver", type=str, default=DEFAULT_PAYMENT_DRIVER)
@async_golem_wrapper
async def allocation_new(
    golem: GolemNode,
    amount: float,
    network: str,
    driver: str,
) -> None:
    allocation = await golem.create_allocation(
        amount,
        network=network,
        driver=driver,
        autoclose=False,
    )
    click.echo(format_allocations([allocation]))


@allocation.command("clean")
@async_golem_wrapper
async def allocation_clean(golem: GolemNode) -> None:
    for allocation in await golem.allocations():
        await allocation.release()
        click.echo(allocation.id)


@cli.command()
@async_golem_wrapper
async def status(golem: GolemNode) -> None:
    allocations = await golem.allocations()
    demands = await golem.demands()
    msg_parts = [
        "ALLOCATIONS",
        format_allocations(allocations),
        "",
        "DEMANDS",
        format_demands(demands),
    ]
    click.echo("\n".join(msg_parts))


@cli.command()
@click.option("--runtime", type=str, required=True)
@click.option("--subnet", type=str, default=DEFAULT_SUBNET)
@click.option("--timeout", "timeout_str", type=str, required=False)
@async_golem_wrapper
async def find_node(
    golem: GolemNode, runtime: str, subnet: str, timeout_str: Optional[str]
) -> None:
    timeout = parse_timedelta_str(timeout_str) if timeout_str is not None else None
    cnt = 0

    async def get_nodes() -> None:
        nonlocal cnt

        payload = CliPayload(runtime)
        demand = await golem.create_demand(payload, subnet=subnet)

        async for proposal in demand.initial_proposals():
            click.echo(format_proposals([proposal], cnt == 0))
            cnt += 1

    try:
        await asyncio.wait_for(get_nodes(), timeout)
    except asyncio.TimeoutError:
        pass
    finally:
        print(f"{cnt} proposals found")



import logging
from typing import Awaitable, Callable, List

from golem.managers.base import Work, WorkManager, WorkResult
from golem.managers.work.mixins import WorkManagerDoWorkMixin, WorkManagerPluginsMixin
from golem.node import GolemNode
from golem.resources import Activity
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


class SequentialWorkManager(WorkManagerPluginsMixin, WorkManagerDoWorkMixin, WorkManager):
    def __init__(
        self, golem: GolemNode, get_activity: Callable[[], Awaitable[Activity]], *args, **kwargs
    ):
        self._get_activity = get_activity

        super().__init__(*args, **kwargs)

    @trace_span("Doing work", show_arguments=True, show_results=True, log_level=logging.INFO)
    async def do_work(self, work: Work) -> WorkResult:
        return await self._do_work_with_plugins(self._do_work, work)

    @trace_span("Doing work list", show_arguments=True, show_results=True, log_level=logging.INFO)
    async def do_work_list(self, work_list: List[Work]) -> List[WorkResult]:
        results = []

        for i, work in enumerate(work_list):
            results.append(await self.do_work(work))

        return results



import asyncio
import re
from dataclasses import MISSING, dataclass
from datetime import datetime
from functools import wraps
from typing import Awaitable, Callable, List, Optional, TypeVar

from prettytable import PrettyTable
from typing_extensions import Concatenate, ParamSpec

from golem.node import GolemNode
from golem.payload import Payload, constraint
from golem.payload.defaults import PROP_RUNTIME_NAME
from golem.resources import Allocation, Demand, Proposal


def format_allocations(allocations: List[Allocation]) -> str:
    x = PrettyTable()
    x.field_names = ["id", "address", "driver", "network", "total", "remaining", "timeout"]
    for allocation in allocations:
        data = allocation.data
        assert data.payment_platform is not None  # mypy
        network, driver, _ = data.payment_platform.split("-")
        x.add_row(
            [
                allocation.id,
                data.address,
                network,
                driver,
                data.total_amount,
                data.remaining_amount,
                data.timeout.isoformat(" ", "seconds") if data.timeout is not None else "",
            ]
        )

    return x.get_string()


def format_demands(demands: List[Demand]) -> str:
    x = PrettyTable()
    x.field_names = ["id", "subnet", "created"]
    for demand in demands:
        data = demand.data
        subnet = data.properties["golem.node.debug.subnet"]

        #   According to ya_client spec, this should be ya_market.models.Timestamp, but is datetime
        #   Maybe this is a TODO for ya_client?
        timestamp: datetime = data.timestamp  # type: ignore

        created = timestamp.isoformat(" ", "seconds")
        x.add_row(
            [
                demand.id,
                subnet,
                created,
            ]
        )
    return x.get_string()


def format_proposals(proposals: List[Proposal], first: bool) -> str:
    x = PrettyTable()
    x.field_names = ["provider_id", "arch", "cores", "threads", "memory (GiB)", "storage (GiB)"]
    for proposal in proposals:
        data = proposal.data
        x.add_row(
            [
                data.issuer_id,
                data.properties["golem.inf.cpu.architecture"],
                data.properties["golem.inf.cpu.cores"],
                data.properties["golem.inf.cpu.threads"],
                round(data.properties["golem.inf.mem.gib"]),
                round(data.properties["golem.inf.storage.gib"]),
            ]
        )

    #   NOTE: this is a "dynamic" table and first row has header and others
    #   have only data.
    lines = x.get_string().splitlines()
    if first:
        return "\n".join(lines[:-1])
    else:
        return lines[3]


@dataclass
class CliPayload(Payload):
    runtime: str = constraint(PROP_RUNTIME_NAME, default=MISSING)


def parse_timedelta_str(timedelta_str: str) -> float:
    """Parse timedelta from string.

    Accepted formats: [float_or_int][s|m|h|d].
    """
    #   TODO: make this compatible with some standard (e.g. "123sec" should be ok maybe?)

    regexp = r"^(\d+|\d+\.\d+)([smhd])?$"
    match = re.search(regexp, timedelta_str)
    if not match:
        raise ValueError("timeout doesn't match the expected format")

    num = float(match.groups()[0])
    what = match.groups()[1] or "s"

    as_seconds = {"s": 1, "m": 60, "h": 60 * 60, "d": 60 * 60 * 24}
    return num * as_seconds[what]


P = ParamSpec("P")
R = TypeVar("R")


def async_golem_wrapper(
    f: Callable[Concatenate[GolemNode, P], Awaitable[R]]
) -> Callable[P, Optional[R]]:
    """Wrap an async function and return a sync function.

    Async function will:

    * start a GolemNode and passes it as the first argument
    * execute the coroutine in a loop
    * on KeyboardInterrupt stop the task and returns None
    """

    @wraps(f)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Optional[R]:
        async def with_golem_node() -> R:
            async with GolemNode(collect_payment_events=False) as golem:
                return await f(golem, *args, **kwargs)

        loop = asyncio.get_event_loop()
        try:
            task = loop.create_task(with_golem_node())
            return loop.run_until_complete(task)
        except KeyboardInterrupt:
            try:
                task.cancel()
                loop.run_until_complete(task)
            except asyncio.CancelledError:
                pass
            return None

    return wrapper



import logging
from typing import Awaitable, Callable

from golem.managers.activity.mixins import ActivityPrepareReleaseMixin, ActivityWrapper
from golem.managers.base import ActivityManager
from golem.node import GolemNode
from golem.resources import Activity, Agreement
from golem.utils.logging import trace_span

logger = logging.getLogger(__name__)


@Activity.register
class SingleUseActivity(ActivityWrapper):
    def __init__(
        self,
        activity,
        release_activity_func,
    ) -> None:
        super().__init__(activity)
        self._release_activity_func = release_activity_func

    async def destroy(self) -> None:
        await self._release_activity_func(self._activity)


class SingleUseActivityManager(ActivityPrepareReleaseMixin, ActivityManager):
    def __init__(
        self, golem: GolemNode, get_agreement: Callable[[], Awaitable[Agreement]], *args, **kwargs
    ):
        self._get_agreement = get_agreement
        self._event_bus = golem.event_bus

        super().__init__(*args, **kwargs)

    @trace_span(show_arguments=True, show_results=True)
    async def get_activity(self) -> Activity:
        while True:
            agreement = await self._get_agreement()
            try:
                activity = await self._prepare_activity(agreement)

                logger.info(f"Activity `{activity}` created")
                # mypy doesn't support `ABCMeta.register` https://github.com/python/mypy/issues/2922
                return SingleUseActivity(
                    activity, self._release_activity
                )  # type: ignore[return-value]

            except Exception:
                logger.exception("Creating activity failed, but will be retried with new agreement")



import json
import logging
import re
from datetime import timedelta
from typing import Callable, Optional

from ya_market import ApiException

from golem.managers import ProposalManagerPlugin
from golem.resources import Proposal
from golem.utils.asyncio.buffer import BackgroundFillBuffer, Buffer, ExpirableBuffer, SimpleBuffer
from golem.utils.asyncio.tasks import resolve_maybe_awaitable
from golem.utils.logging import trace_span
from golem.utils.typing import MaybeAwaitable

logger = logging.getLogger(__name__)


class ProposalBuffer(ProposalManagerPlugin):
    def __init__(
        self,
        min_size: int,
        max_size: int,
        fill_concurrency_size: int = 1,
        fill_at_start=False,
        get_expiration_func: Optional[
            Callable[[Proposal], MaybeAwaitable[Optional[timedelta]]]
        ] = None,
        on_expiration_func: Optional[Callable[[Proposal], MaybeAwaitable[None]]] = None,
    ) -> None:
        self._min_size = min_size
        self._max_size = max_size
        self._fill_concurrency_size = fill_concurrency_size
        self._fill_at_start = fill_at_start
        self._get_expiration_func = get_expiration_func
        self._on_expiration_func = on_expiration_func

        # TODO: Consider moving buffer composition from here to plugin level
        buffer: Buffer[Proposal] = SimpleBuffer()

        if self._get_expiration_func:
            buffer = ExpirableBuffer(
                buffer=buffer,
                get_expiration_func=self._get_expiration_func,
                on_expired_func=self._on_expired,
            )

        self._buffer: BackgroundFillBuffer[Proposal] = BackgroundFillBuffer(
            buffer=buffer,
            fill_func=self._call_feed_func,
            fill_concurrency_size=self._fill_concurrency_size,
            on_added_func=self._on_added,
        )

    async def _call_feed_func(self) -> Proposal:
        return await self._get_proposal()

    async def _on_added(self, proposal: Proposal) -> None:
        count_current = self._buffer.size()
        count_with_requested = self._buffer.size_with_requested()
        pending = count_with_requested - count_current

        logger.debug(
            "Proposal added, having %d proposals, and %d already requested, target %d",
            count_current,
            pending,
            self._max_size,
        )

    async def _on_expired(self, proposal: Proposal):
        logger.debug("Rejecting expired `%r` and requesting fill", proposal)

        try:
            if proposal.draft:
                await proposal.reject("Proposal no longer needed due to its near expiration.")
        except ApiException as e:
            message = json.loads(e.body)["message"]
            if e.status == 400 and re.match(
                r"^Subscription \[([^]]+)\] (wasn't found|expired|was already unsubscribed).$",
                message,
            ):
                logger.warning(
                    f"Proposal assumed already expired. Consider shortening the expiry duration:"
                    f" `{message}`"
                )
            else:
                raise

        await self._request_proposals()

        if self._on_expiration_func:
            await resolve_maybe_awaitable(self._on_expiration_func(proposal))

    @trace_span()
    async def start(self) -> None:
        await self._buffer.start()

        if self._fill_at_start:
            await self._request_proposals()

    @trace_span()
    async def stop(self) -> None:
        await self._buffer.stop()

    async def _request_proposals(self) -> None:
        count_current = self._buffer.size()
        count_with_requested = self._buffer.size_with_requested()
        requested = self._max_size - count_with_requested

        logger.debug(
            "Proposal count %d and %d already requested, requesting additional %d to match"
            " target %d",
            count_current,
            count_with_requested - count_current,
            requested,
            self._max_size,
        )

        await self._buffer.request(requested)

    @trace_span(show_results=True)
    async def get_proposal(self) -> Proposal:
        if not self._get_buffered_proposals_count():
            logger.debug("No proposals to get, requesting fill")
            await self._request_proposals()

        proposal = await self._get_buffered_proposal()

        proposals_count = self._get_buffered_proposals_count()
        if proposals_count < self._min_size:
            logger.debug(
                "Proposals count %d is below minimum size %d, requesting fill",
                proposals_count,
                self._min_size,
            )
            await self._request_proposals()
        else:
            logger.debug(
                "Proposals count %d is above minimum size %d, skipping fill",
                proposals_count,
                self._min_size,
            )

        return proposal

    async def _get_buffered_proposal(self) -> Proposal:
        return await self._buffer.get()

    def _get_buffered_proposals_count(self) -> int:
        return self._buffer.size()
